/*
 *  [The "BSD license"]
 *  Copyright (c) 2002-2011, Rodney O'Donnell, Lloyd Allison, Kevin Korb
 *  Copyright (c) 2002-2011, Monash University
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *    1. Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *    2. Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *    3. The name of the author may not be used to endorse or promote products
 *       derived from this software without specific prior written permission.*
 *
 *  THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 *  IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 *  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package camml.core.newgui;

import java.awt.Color;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.OutputStream;
import java.io.PrintStream;
import java.util.Random;

import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JOptionPane;
import javax.swing.JTable;
import javax.swing.JTextArea;
import javax.swing.UIManager;
import javax.swing.UnsupportedLookAndFeelException;
import javax.swing.filechooser.FileNameExtensionFilter;


/**Main class for GUI.
 * Most GUI-related code is in this class.<br>
 * Most CaMML-related code is in the GUIModel class.<br>
 * Also see:<br>
 * - GUIParameters for some options. <br>
 * - GUIModel for the code that runs search, exports results, etc.<br>
 * - Start the GUI via the RunGUI class.<br>
 * 
 * Note: GUI created with NetBeans IDE (using Java's Swing GUI libraries); much
 * 	of the code (especially in initComponents()) is generated by NetBeans.<br>
 * GUI is presently set up to use the same visual style ("Look and Feel") as the
 * operating system that it is currently being run on.
 * @author Alex Black
 */
public class CaMMLGUI extends javax.swing.JFrame {
	private static final long serialVersionUID = 2179376424056735519L;
	
	private GUIModel guimodel;			//Reference to the object that actually runs search, exports BNs etc
	
    /**Constructor. Creates a CaMMLGUI object.
     */
    public CaMMLGUI( GUIModel model ) {
    	
    	this.guimodel = model;
    	
        //Set look and feel:
    	try {
            // Set System L&F
	        UIManager.setLookAndFeel(
	            UIManager.getSystemLookAndFeelClassName());
	    } 
	    catch (UnsupportedLookAndFeelException e) {
	       // handle exception
	    }
	    catch (ClassNotFoundException e) {
	       // handle exception
	    }
	    catch (InstantiationException e) {
	       // handle exception
	    }
	    catch (IllegalAccessException e) {
	       // handle exception
	    }
        
    	//Initialise Swing GUI Components:
        initComponents();
        
        //Redirect console output to GUI text area:
        redirectOutputStream( runConsoleOutputTextArea );
        
        //Expert priors disabled by default (until user selects the expert priors checkbox):
        setExpertPriorsTabEnabled( false );
        
        //Populate the options available to the user in setup tab: (from options in GUIParameters)
        populateSetupOptions();   
    }

    /**
     * This method is called from within the constructor to initialize the form.
     * Most of this code is auto-generated by NetBeans (except at very end)
     */
    private void initComponents() {

    	//--------- Start NetBeans Autogenerated Code ---------
    	
    	mainTabbedPane = new javax.swing.JTabbedPane();
        setupPanel = new javax.swing.JPanel();
        jLabelDataFileLabel = new javax.swing.JLabel();
        filePathTextField = new javax.swing.JTextField();
        viewDataBtn = new javax.swing.JButton();
        loadDataFileBtn = new javax.swing.JButton();
        searchParametersScrollPane = new javax.swing.JScrollPane();
        searchParametersPanel = new javax.swing.JPanel();
        setupLabelMMLLearner = new javax.swing.JLabel();
        MMLLearnerCombo = new javax.swing.JComboBox();
        setupLabelSearchFactor = new javax.swing.JLabel();
        searchFactorTextfield = new javax.swing.JTextField();
        setupLabelMinTotalPosterior = new javax.swing.JLabel();
        setupLabelMaxSECs = new javax.swing.JLabel();
        maxSECsTextfield = new javax.swing.JTextField();
        setupLabelParameterizationSettings = new javax.swing.JLabel();
        jSeparator3 = new javax.swing.JSeparator();
        setupLabelSearchSettings = new javax.swing.JLabel();
        minTotalPosteriorTextField = new javax.swing.JTextField();
        setupLabelRNG = new javax.swing.JLabel();
        setupRNGCombo = new javax.swing.JComboBox();
        setupLabelRNGSeed = new javax.swing.JLabel();
        setupRNGSeed = new javax.swing.JTextField();
        jSeparator4 = new javax.swing.JSeparator();
        setupLabelAdvancedSettings = new javax.swing.JLabel();
        setupLabelLearnDBN = new javax.swing.JLabel();
        setupLearnDBNCheckbox = new javax.swing.JCheckBox();
        setupRNGSeed2 = new javax.swing.JTextField();
        priorsPanel = new javax.swing.JPanel();
        useExpertPriorsCheckbox = new javax.swing.JCheckBox();
        saveExpertPriorsBtn = new javax.swing.JButton();
        loadExpertPriorsBtn = new javax.swing.JButton();
        newExpertPriorsBtn = new javax.swing.JButton();
        jScrollPane1 = new javax.swing.JScrollPane();
        expertPriorsTextArea = new javax.swing.JTextArea();
        validateExpertPriorsBtn = new javax.swing.JButton();
        runPanel = new javax.swing.JPanel();
        runCammlRunBtn = new javax.swing.JButton();
        jScrollPane2 = new javax.swing.JScrollPane();
        runConsoleOutputTextArea = new javax.swing.JTextArea();
        runLabelConsoleOutput = new javax.swing.JLabel();
        runClearOutputBtn = new javax.swing.JButton();
        resultsPanel = new javax.swing.JPanel();
        resultsLabelFullResults = new javax.swing.JLabel();
        resultsScrollPane = new javax.swing.JScrollPane();
        resultsExportSelectedBN = new javax.swing.JButton();
        resultsExportAllNetworks = new javax.swing.JButton();
        resultsViewSelectedNetworkBtn = new javax.swing.JButton();
        statusBarLabelData = new javax.swing.JLabel();
        versionLabel = new javax.swing.JLabel();
        statusBarLabelExpertPriors = new javax.swing.JLabel();
        jMenuBar1 = new javax.swing.JMenuBar();
        fileMenu = new javax.swing.JMenu();
        exitMenuItem = new javax.swing.JMenuItem();
        helpMenu = new javax.swing.JMenu();
        helpMenuItem = new javax.swing.JMenuItem();
        jSeparator2 = new javax.swing.JPopupMenu.Separator();
        aboutMenuItem = new javax.swing.JMenuItem();

        setDefaultCloseOperation(javax.swing.WindowConstants.DO_NOTHING_ON_CLOSE);
        setTitle("CaMML");
        setMinimumSize(new java.awt.Dimension(640, 320));
        addWindowListener(new java.awt.event.WindowAdapter() {
            public void windowClosing(java.awt.event.WindowEvent evt) {
                formWindowClosing(evt);
            }
        });

        jLabelDataFileLabel.setText("Data File:");

        filePathTextField.setEditable(false);

        viewDataBtn.setLabel("View");
        viewDataBtn.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                viewDataBtnActionPerformed(evt);
            }
        });

        loadDataFileBtn.setText("Load");
        loadDataFileBtn.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                loadDataFileBtnActionPerformed(evt);
            }
        });

        searchParametersScrollPane.setBorder(javax.swing.BorderFactory.createTitledBorder("Search Parameters"));
        searchParametersScrollPane.setHorizontalScrollBarPolicy(javax.swing.ScrollPaneConstants.HORIZONTAL_SCROLLBAR_NEVER);

        searchParametersPanel.setPreferredSize(new java.awt.Dimension(633, 280));

        setupLabelMMLLearner.setText("Parameterization Type");
        setupLabelMMLLearner.setToolTipText("<html>Parameterization type for Bayesian Network nodes.<br>\nBasically: What sort of distribution should be used?<br>\n- CPT = Conditional Probability Table<br>\n- DTree = Decision Tree<br>\n- Logit = Logit Distribution</html>");

        MMLLearnerCombo.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                MMLLearnerComboActionPerformed(evt);
            }
        });

        setupLabelSearchFactor.setText("Search Factor");
        setupLabelSearchFactor.setToolTipText("<html>Search factor multiplies the base number of loops of the search algorithm. Default: 1.0<br>\nHigher numbers mean a greater number of loops in the Metropolis Algorithm are<br>\nrun, with a corresponding increase in the amount of time taken for the algorithm to complete.<br>\nFor example: Search Factor = 2.0 -> Twice as many model changes attempted.</html>");

        searchFactorTextfield.setHorizontalAlignment(javax.swing.JTextField.RIGHT);
        searchFactorTextfield.setText("1.0");
        searchFactorTextfield.addFocusListener(new java.awt.event.FocusAdapter() {
            public void focusLost(java.awt.event.FocusEvent evt) {
                searchFactorTextfieldFocusLost(evt);
            }
        });

        setupLabelMinTotalPosterior.setText("Minimum Total Posterior");
        setupLabelMinTotalPosterior.setToolTipText("<html>Minimal total SEC posterior to retain from Metropolis Sampling.<br>\nNote: This value is ignored if more than the specified 'Maximum Number<br>\nof SECs' is required.</html>");

        setupLabelMaxSECs.setText("Maximum Number of SECs");
        setupLabelMaxSECs.setToolTipText("<html>Maximum number of SECs to retain post-Metropolis sampling.<br>\nNote: Making this number larger slows down the process of combining<br>\nmodels after running search (i.e time taken to generate full results).</html>");

        maxSECsTextfield.setHorizontalAlignment(javax.swing.JTextField.RIGHT);
        maxSECsTextfield.setText("30");
        maxSECsTextfield.addFocusListener(new java.awt.event.FocusAdapter() {
            public void focusLost(java.awt.event.FocusEvent evt) {
                maxSECsTextfieldFocusLost(evt);
            }
        });

        setupLabelParameterizationSettings.setFont(new java.awt.Font("Tahoma", 1, 11)); // NOI18N
        setupLabelParameterizationSettings.setText("Bayesian Network Parameterization Settings");

        setupLabelSearchSettings.setFont(new java.awt.Font("Tahoma", 1, 11)); // NOI18N
        setupLabelSearchSettings.setText("Metropolis Algorithm Search Settings");

        minTotalPosteriorTextField.setHorizontalAlignment(javax.swing.JTextField.RIGHT);
        minTotalPosteriorTextField.setText("0.999");
        minTotalPosteriorTextField.addFocusListener(new java.awt.event.FocusAdapter() {
            public void focusLost(java.awt.event.FocusEvent evt) {
                minTotalPosteriorTextFieldFocusLost(evt);
            }
        });

        setupLabelRNG.setText("Random Number Generator");
        setupLabelRNG.setToolTipText("<html>Random Number Generator used during search.<br>\nIn general, the default RNG is appropriate, unless there is a specific<br>\nreason (i.e. regression testing, etc) to use one of the others.\n</html>\n");

        setupRNGCombo.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                setupRNGComboActionPerformed(evt);
            }
        });

        setupLabelRNGSeed.setText("Random Number Generator Seed/s");
        setupLabelRNGSeed.setToolTipText("<html>Seed values for the random number generators.<br>\nText boxes are only enabled if a RNG that allows for the seed to<br>\nbe set is selected in the Random Number Generator combo box.</html>");

        setupRNGSeed.setHorizontalAlignment(javax.swing.JTextField.RIGHT);
        setupRNGSeed.addFocusListener(new java.awt.event.FocusAdapter() {
            public void focusLost(java.awt.event.FocusEvent evt) {
                setupRNGSeedFocusLost(evt);
            }
        });

        setupLabelAdvancedSettings.setFont(new java.awt.Font("Tahoma", 1, 11)); // NOI18N
        setupLabelAdvancedSettings.setText("Advanced Settings");

        setupLabelLearnDBN.setText("Learn Dynamic Bayesian Network (DBN)");
        setupLabelLearnDBN.setToolTipText("<html>CaMML can learn two types of Bayesian Networks:<br>\n- A standard Bayesian Network, or<br>\n- A Dynamic Bayesian Network (DBN)<br>\nIf checked, CaMML will assume the selected data file represents a time series,<br>\nand will attempt to learn a DBN from this time series.</html>");

        setupLearnDBNCheckbox.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                setupLearnDBNCheckboxActionPerformed(evt);
            }
        });

        setupRNGSeed2.addFocusListener(new java.awt.event.FocusAdapter() {
            public void focusLost(java.awt.event.FocusEvent evt) {
                setupRNGSeed2FocusLost(evt);
            }
        });

        javax.swing.GroupLayout searchParametersPanelLayout = new javax.swing.GroupLayout(searchParametersPanel);
        searchParametersPanel.setLayout(searchParametersPanelLayout);
        searchParametersPanelLayout.setHorizontalGroup(
            searchParametersPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addComponent(jSeparator3)
            .addGroup(searchParametersPanelLayout.createSequentialGroup()
                .addContainerGap()
                .addGroup(searchParametersPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
                    .addGroup(searchParametersPanelLayout.createSequentialGroup()
                        .addComponent(setupLabelParameterizationSettings)
                        .addContainerGap())
                    .addGroup(searchParametersPanelLayout.createSequentialGroup()
                        .addGroup(searchParametersPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
                            .addComponent(setupLabelMMLLearner, javax.swing.GroupLayout.PREFERRED_SIZE, 300, javax.swing.GroupLayout.PREFERRED_SIZE)
                            .addComponent(setupLabelSearchSettings)
                            .addGroup(searchParametersPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.TRAILING, false)
                                .addComponent(setupLabelRNG, javax.swing.GroupLayout.Alignment.LEADING, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, Short.MAX_VALUE)
                                .addComponent(setupLabelRNGSeed, javax.swing.GroupLayout.Alignment.LEADING, javax.swing.GroupLayout.DEFAULT_SIZE, 211, Short.MAX_VALUE)
                                .addComponent(setupLabelSearchFactor, javax.swing.GroupLayout.Alignment.LEADING, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, Short.MAX_VALUE)
                                .addComponent(setupLabelMaxSECs, javax.swing.GroupLayout.Alignment.LEADING, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, Short.MAX_VALUE)
                                .addComponent(setupLabelMinTotalPosterior, javax.swing.GroupLayout.Alignment.LEADING, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, Short.MAX_VALUE))
                            .addComponent(setupLabelAdvancedSettings)
                            .addComponent(setupLabelLearnDBN, javax.swing.GroupLayout.PREFERRED_SIZE, 211, javax.swing.GroupLayout.PREFERRED_SIZE))
                        .addGap(12, 12, 12)
                        .addGroup(searchParametersPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
                            .addGroup(searchParametersPanelLayout.createSequentialGroup()
                                .addGroup(searchParametersPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING, false)
                                    .addComponent(setupLearnDBNCheckbox)
                                    .addComponent(maxSECsTextfield)
                                    .addComponent(searchFactorTextfield)
                                    .addComponent(minTotalPosteriorTextField, javax.swing.GroupLayout.DEFAULT_SIZE, 120, Short.MAX_VALUE))
                                .addGap(0, 0, Short.MAX_VALUE))
                            .addGroup(searchParametersPanelLayout.createSequentialGroup()
                                .addGroup(searchParametersPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.TRAILING, false)
                                    .addGroup(javax.swing.GroupLayout.Alignment.LEADING, searchParametersPanelLayout.createSequentialGroup()
                                        .addComponent(setupRNGSeed, javax.swing.GroupLayout.PREFERRED_SIZE, 120, javax.swing.GroupLayout.PREFERRED_SIZE)
                                        .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                                        .addComponent(setupRNGSeed2, javax.swing.GroupLayout.PREFERRED_SIZE, 120, javax.swing.GroupLayout.PREFERRED_SIZE))
                                    .addComponent(setupRNGCombo, javax.swing.GroupLayout.Alignment.LEADING, 0, javax.swing.GroupLayout.DEFAULT_SIZE, Short.MAX_VALUE)
                                    .addComponent(MMLLearnerCombo, javax.swing.GroupLayout.Alignment.LEADING, javax.swing.GroupLayout.PREFERRED_SIZE, 266, javax.swing.GroupLayout.PREFERRED_SIZE))
                                .addContainerGap(45, Short.MAX_VALUE))))))
            .addComponent(jSeparator4)
        );
        searchParametersPanelLayout.setVerticalGroup(
            searchParametersPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGroup(searchParametersPanelLayout.createSequentialGroup()
                .addContainerGap()
                .addComponent(setupLabelParameterizationSettings)
                .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                .addGroup(searchParametersPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.BASELINE)
                    .addComponent(setupLabelMMLLearner)
                    .addComponent(MMLLearnerCombo, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE))
                .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                .addComponent(jSeparator3, javax.swing.GroupLayout.PREFERRED_SIZE, 10, javax.swing.GroupLayout.PREFERRED_SIZE)
                .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                .addGroup(searchParametersPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.TRAILING)
                    .addGroup(searchParametersPanelLayout.createSequentialGroup()
                        .addComponent(setupLabelSearchSettings)
                        .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                        .addGroup(searchParametersPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.BASELINE)
                            .addComponent(setupLabelRNG)
                            .addComponent(setupRNGCombo, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE))
                        .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                        .addGroup(searchParametersPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.BASELINE)
                            .addComponent(setupLabelRNGSeed)
                            .addComponent(setupRNGSeed, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE)
                            .addComponent(setupRNGSeed2, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE))
                        .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                        .addGroup(searchParametersPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.BASELINE)
                            .addComponent(setupLabelSearchFactor)
                            .addComponent(searchFactorTextfield, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE))
                        .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                        .addGroup(searchParametersPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.BASELINE)
                            .addComponent(setupLabelMaxSECs)
                            .addComponent(maxSECsTextfield, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE))
                        .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                        .addGroup(searchParametersPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.BASELINE)
                            .addComponent(setupLabelMinTotalPosterior)
                            .addComponent(minTotalPosteriorTextField, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE))
                        .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                        .addComponent(jSeparator4, javax.swing.GroupLayout.PREFERRED_SIZE, 10, javax.swing.GroupLayout.PREFERRED_SIZE)
                        .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                        .addComponent(setupLabelAdvancedSettings)
                        .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                        .addComponent(setupLabelLearnDBN))
                    .addComponent(setupLearnDBNCheckbox))
                .addContainerGap(74, Short.MAX_VALUE))
        );

        setupLabelLearnDBN.getAccessibleContext().setAccessibleName("Learn DBN");

        searchParametersScrollPane.setViewportView(searchParametersPanel);

        javax.swing.GroupLayout setupPanelLayout = new javax.swing.GroupLayout(setupPanel);
        setupPanel.setLayout(setupPanelLayout);
        setupPanelLayout.setHorizontalGroup(
            setupPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGroup(setupPanelLayout.createSequentialGroup()
                .addContainerGap()
                .addGroup(setupPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
                    .addComponent(searchParametersScrollPane, javax.swing.GroupLayout.DEFAULT_SIZE, 606, Short.MAX_VALUE)
                    .addGroup(setupPanelLayout.createSequentialGroup()
                        .addComponent(jLabelDataFileLabel, javax.swing.GroupLayout.PREFERRED_SIZE, 48, javax.swing.GroupLayout.PREFERRED_SIZE)
                        .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                        .addComponent(filePathTextField)
                        .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                        .addComponent(loadDataFileBtn)
                        .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                        .addComponent(viewDataBtn)))
                .addContainerGap())
        );
        setupPanelLayout.setVerticalGroup(
            setupPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGroup(setupPanelLayout.createSequentialGroup()
                .addContainerGap()
                .addGroup(setupPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.BASELINE)
                    .addComponent(jLabelDataFileLabel)
                    .addComponent(filePathTextField, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE)
                    .addComponent(viewDataBtn)
                    .addComponent(loadDataFileBtn))
                .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                .addComponent(searchParametersScrollPane, javax.swing.GroupLayout.DEFAULT_SIZE, 370, Short.MAX_VALUE))
        );

        mainTabbedPane.addTab("Setup", setupPanel);

        useExpertPriorsCheckbox.setText("Use Expert Priors");
        useExpertPriorsCheckbox.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                useExpertPriorsCheckboxActionPerformed(evt);
            }
        });

        saveExpertPriorsBtn.setText("Save");
        saveExpertPriorsBtn.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                saveExpertPriorsBtnActionPerformed(evt);
            }
        });

        loadExpertPriorsBtn.setText("Load");
        loadExpertPriorsBtn.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                loadExpertPriorsBtnActionPerformed(evt);
            }
        });

        newExpertPriorsBtn.setText("New");
        newExpertPriorsBtn.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                newExpertPriorsBtnActionPerformed(evt);
            }
        });

        expertPriorsTextArea.setColumns(20);
        expertPriorsTextArea.setFont(new java.awt.Font("Monospaced", 0, 12)); // NOI18N
        expertPriorsTextArea.setRows(5);
        expertPriorsTextArea.setEnabled(false);
        expertPriorsTextArea.addFocusListener(new java.awt.event.FocusAdapter() {
            public void focusLost(java.awt.event.FocusEvent evt) {
                expertPriorsTextAreaFocusLost(evt);
            }
        });
        jScrollPane1.setViewportView(expertPriorsTextArea);

        validateExpertPriorsBtn.setText("Validate");
        validateExpertPriorsBtn.setEnabled(false);
        validateExpertPriorsBtn.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                validateExpertPriorsBtnActionPerformed(evt);
            }
        });

        javax.swing.GroupLayout priorsPanelLayout = new javax.swing.GroupLayout(priorsPanel);
        priorsPanel.setLayout(priorsPanelLayout);
        priorsPanelLayout.setHorizontalGroup(
            priorsPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGroup(priorsPanelLayout.createSequentialGroup()
                .addContainerGap()
                .addGroup(priorsPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
                    .addComponent(jScrollPane1)
                    .addGroup(priorsPanelLayout.createSequentialGroup()
                        .addComponent(useExpertPriorsCheckbox)
                        .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED, 320, Short.MAX_VALUE)
                        .addComponent(newExpertPriorsBtn)
                        .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                        .addComponent(loadExpertPriorsBtn)
                        .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                        .addComponent(saveExpertPriorsBtn))
                    .addGroup(javax.swing.GroupLayout.Alignment.TRAILING, priorsPanelLayout.createSequentialGroup()
                        .addGap(0, 0, Short.MAX_VALUE)
                        .addComponent(validateExpertPriorsBtn)))
                .addContainerGap())
        );
        priorsPanelLayout.setVerticalGroup(
            priorsPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGroup(priorsPanelLayout.createSequentialGroup()
                .addContainerGap()
                .addGroup(priorsPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.BASELINE)
                    .addComponent(useExpertPriorsCheckbox)
                    .addComponent(saveExpertPriorsBtn)
                    .addComponent(loadExpertPriorsBtn)
                    .addComponent(newExpertPriorsBtn))
                .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                .addComponent(jScrollPane1, javax.swing.GroupLayout.DEFAULT_SIZE, 341, Short.MAX_VALUE)
                .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                .addComponent(validateExpertPriorsBtn))
        );

        mainTabbedPane.addTab("Expert Priors", priorsPanel);

        runCammlRunBtn.setText("Run CaMML");
        runCammlRunBtn.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                runCammlRunBtnActionPerformed(evt);
            }
        });

        runConsoleOutputTextArea.setColumns(20);
        runConsoleOutputTextArea.setFont(new java.awt.Font("Monospaced", 0, 12)); // NOI18N
        runConsoleOutputTextArea.setRows(5);
        jScrollPane2.setViewportView(runConsoleOutputTextArea);

        runLabelConsoleOutput.setText("Console Output:");

        runClearOutputBtn.setText("Clear Output");
        runClearOutputBtn.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                runClearOutputBtnActionPerformed(evt);
            }
        });

        javax.swing.GroupLayout runPanelLayout = new javax.swing.GroupLayout(runPanel);
        runPanel.setLayout(runPanelLayout);
        runPanelLayout.setHorizontalGroup(
            runPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGroup(runPanelLayout.createSequentialGroup()
                .addContainerGap()
                .addGroup(runPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
                    .addComponent(jScrollPane2, javax.swing.GroupLayout.DEFAULT_SIZE, 606, Short.MAX_VALUE)
                    .addGroup(runPanelLayout.createSequentialGroup()
                        .addComponent(runLabelConsoleOutput)
                        .addGap(0, 0, Short.MAX_VALUE))
                    .addGroup(runPanelLayout.createSequentialGroup()
                        .addComponent(runCammlRunBtn)
                        .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED, 422, Short.MAX_VALUE)
                        .addComponent(runClearOutputBtn)))
                .addContainerGap())
        );
        runPanelLayout.setVerticalGroup(
            runPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGroup(runPanelLayout.createSequentialGroup()
                .addContainerGap()
                .addGroup(runPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.BASELINE)
                    .addComponent(runCammlRunBtn)
                    .addComponent(runClearOutputBtn))
                .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                .addComponent(runLabelConsoleOutput)
                .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                .addComponent(jScrollPane2, javax.swing.GroupLayout.DEFAULT_SIZE, 339, Short.MAX_VALUE)
                .addContainerGap())
        );

        mainTabbedPane.addTab("Run", runPanel);

        resultsLabelFullResults.setFont(new java.awt.Font("Tahoma", 1, 11)); // NOI18N
        resultsLabelFullResults.setText("Results");

        resultsScrollPane.setBorder(javax.swing.BorderFactory.createTitledBorder("Learned Networks"));

        resultsExportSelectedBN.setText("Export Selected Network");
        resultsExportSelectedBN.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                resultsExportSelectedBNActionPerformed(evt);
            }
        });

        resultsExportAllNetworks.setText("Export All");
        resultsExportAllNetworks.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                resultsExportAllNetworksActionPerformed(evt);
            }
        });

        resultsViewSelectedNetworkBtn.setText("View Selected Network");
        resultsViewSelectedNetworkBtn.setMaximumSize(new java.awt.Dimension(151, 23));
        resultsViewSelectedNetworkBtn.setMinimumSize(new java.awt.Dimension(151, 23));
        resultsViewSelectedNetworkBtn.setPreferredSize(new java.awt.Dimension(151, 23));
        resultsViewSelectedNetworkBtn.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                resultsViewSelectedNetworkBtnActionPerformed(evt);
            }
        });

        javax.swing.GroupLayout resultsPanelLayout = new javax.swing.GroupLayout(resultsPanel);
        resultsPanel.setLayout(resultsPanelLayout);
        resultsPanelLayout.setHorizontalGroup(
            resultsPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGroup(resultsPanelLayout.createSequentialGroup()
                .addContainerGap()
                .addGroup(resultsPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
                    .addComponent(resultsScrollPane)
                    .addGroup(resultsPanelLayout.createSequentialGroup()
                        .addComponent(resultsViewSelectedNetworkBtn, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE)
                        .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.UNRELATED)
                        .addComponent(resultsExportSelectedBN, javax.swing.GroupLayout.PREFERRED_SIZE, 151, javax.swing.GroupLayout.PREFERRED_SIZE)
                        .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED, 215, Short.MAX_VALUE)
                        .addComponent(resultsExportAllNetworks))
                    .addGroup(resultsPanelLayout.createSequentialGroup()
                        .addComponent(resultsLabelFullResults)
                        .addGap(0, 0, Short.MAX_VALUE)))
                .addContainerGap())
        );
        resultsPanelLayout.setVerticalGroup(
            resultsPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGroup(resultsPanelLayout.createSequentialGroup()
                .addContainerGap()
                .addComponent(resultsLabelFullResults)
                .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                .addComponent(resultsScrollPane, javax.swing.GroupLayout.DEFAULT_SIZE, 350, Short.MAX_VALUE)
                .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                .addGroup(resultsPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.BASELINE)
                    .addComponent(resultsExportAllNetworks)
                    .addComponent(resultsViewSelectedNetworkBtn, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE)
                    .addComponent(resultsExportSelectedBN)))
        );

        resultsScrollPane.getAccessibleContext().setAccessibleName("Results");

        mainTabbedPane.addTab("Results", resultsPanel);

        statusBarLabelData.setText("Data: Not Loaded");

        versionLabel.setText("CaMML (C) - Ver. 0.00");

        statusBarLabelExpertPriors.setText("Expert Priors: Not Set");

        fileMenu.setText("File");

        exitMenuItem.setText("Exit");
        exitMenuItem.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                exitMenuItemActionPerformed(evt);
            }
        });
        fileMenu.add(exitMenuItem);

        jMenuBar1.add(fileMenu);

        helpMenu.setText("Help");

        helpMenuItem.setText("Help");
        helpMenuItem.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                helpMenuItemActionPerformed(evt);
            }
        });
        helpMenu.add(helpMenuItem);
        helpMenu.add(jSeparator2);

        aboutMenuItem.setText("About");
        aboutMenuItem.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                aboutMenuItemActionPerformed(evt);
            }
        });
        helpMenu.add(aboutMenuItem);

        jMenuBar1.add(helpMenu);

        setJMenuBar(jMenuBar1);

        javax.swing.GroupLayout layout = new javax.swing.GroupLayout(getContentPane());
        getContentPane().setLayout(layout);
        layout.setHorizontalGroup(
            layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addComponent(mainTabbedPane, javax.swing.GroupLayout.PREFERRED_SIZE, 0, Short.MAX_VALUE)
            .addGroup(layout.createSequentialGroup()
                .addContainerGap()
                .addComponent(statusBarLabelData, javax.swing.GroupLayout.PREFERRED_SIZE, 160, javax.swing.GroupLayout.PREFERRED_SIZE)
                .addGap(18, 18, 18)
                .addComponent(statusBarLabelExpertPriors, javax.swing.GroupLayout.PREFERRED_SIZE, 201, javax.swing.GroupLayout.PREFERRED_SIZE)
                .addGap(117, 117, 117)
                .addComponent(versionLabel, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, Short.MAX_VALUE)
                .addGap(18, 18, 18))
        );
        layout.setVerticalGroup(
            layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGroup(layout.createSequentialGroup()
                .addComponent(mainTabbedPane)
                .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.UNRELATED)
                .addGroup(layout.createParallelGroup(javax.swing.GroupLayout.Alignment.BASELINE)
                    .addComponent(statusBarLabelData)
                    .addComponent(versionLabel)
                    .addComponent(statusBarLabelExpertPriors)))
        );

        mainTabbedPane.getAccessibleContext().setAccessibleName("Setup");
        
        
        
        //------------ End NetBeans Generated Code ------------
        //***************************************************
        
        versionLabel.setText("CaMML (C) - Ver. " + GUIParameters.versionNumber );
        
        //Set uf the file dialogue boxes:
        dataFileChooser = new JFileChooser();
        dataFileChooser.setFileFilter(  new FileNameExtensionFilter("Old CaMML Format (.cas)", "cas") );
        dataFileChooser.addChoosableFileFilter( new FileNameExtensionFilter("Friedman Data Files (.data)","data") );	//TODO: Currently untested!
        dataFileChooser.addChoosableFileFilter( new FileNameExtensionFilter("Weka Data Files (.arff)","arff") );
        
        outputBNetChooser = new JFileChooser();
        outputBNetChooser.setFileFilter( new FileNameExtensionFilter("Netica BN (.dne)","dne"));
        
        expertPriorsFileChooser = new JFileChooser();
        expertPriorsFileChooser.setFileFilter( new FileNameExtensionFilter("Text file (.txt)","txt") );

        pack();
    }

    
    
    
    
    //***********************************************************
    /*----------------- Menu and Misc. Actions: ---------------*/
    //***********************************************************
    /**User clicked on File -> Exit menu*/
    private void exitMenuItemActionPerformed(java.awt.event.ActionEvent evt) {
    	confirmBeforeExit();
    }
    
    /**User clicked 'X' on main window*/
    private void formWindowClosing(java.awt.event.WindowEvent evt) {
    	confirmBeforeExit();
    }
    
    private void confirmBeforeExit(){
    	int result = JOptionPane.showConfirmDialog(mainTabbedPane, "Are you sure you want to exit? Any unsaved data will be lost.", "Exit CaMML?", JOptionPane.OK_CANCEL_OPTION);
    	if( result == JOptionPane.OK_OPTION ) System.exit(0);
    }
   
    /**User clicked on Help -> Help*/
    private void helpMenuItemActionPerformed(java.awt.event.ActionEvent evt) {
        // TODO Not yet implemented.
    	JOptionPane.showMessageDialog( mainTabbedPane, "Not implemented.");
    }

    /**User clicked on Help -> About*/
    private void aboutMenuItemActionPerformed(java.awt.event.ActionEvent evt) {
        // TODO Not yet implemented
    	JOptionPane.showMessageDialog( mainTabbedPane, "Not implemented.");
    }
   
    
    
    
    //***********************************************************
    /*------------------- Setup Tab Actions: ------------------*/
    //***********************************************************
    
    /**User clicked on 'Load' (i.e. load data) in Setup tab*/
    private void loadDataFileBtnActionPerformed(java.awt.event.ActionEvent evt) {
        
    	//Set directory:
    	if( guimodel.selectedFile == null ){
    		dataFileChooser.setCurrentDirectory( new File(".") );
    	} else{
    		dataFileChooser.setCurrentDirectory( guimodel.selectedFile );
    	}
    	
    	//Open the file dialog box; work out what the user actually did (OK, cancel etc)
    	int result = dataFileChooser.showOpenDialog(this);
    	
    	if( result == JFileChooser.APPROVE_OPTION ){	//If the user selects a file to open 
    		guimodel.selectedFile = dataFileChooser.getSelectedFile();
    		try{
    			guimodel.loadDataFile( dataFileChooser.getSelectedFile() );
    			filePathTextField.setText( dataFileChooser.getSelectedFile().getAbsolutePath() );
    		} catch( Exception e ){
    			JOptionPane.showMessageDialog(mainTabbedPane, "Error loading data\n" + e.getMessage() );
    			filePathTextField.setText( "" );
    			updateStatusBarLabelData();
    			return;
    		}
    	}
    	
    	//Check if data is continuous...
    	if( !guimodel.checkDataNotContinuous() ){
    		JOptionPane.showMessageDialog(mainTabbedPane, "Dataset contains continuous variables. Handling of continuous datasets not yet implemented." );
    		guimodel.data = null;
    		filePathTextField.setText("");
    	}
    	updateStatusBarLabelData();
    }
    
    /**User clicked on 'View' (i.e. view data) in Setup tab.
     * Creates a DataViewer object and displays it. (See: DataViewer class)
     */
    private void viewDataBtnActionPerformed(java.awt.event.ActionEvent evt) {
    	createDataViewer();
    }
    
    /**User selects a different option (i.e. CPT, DTree, etc) from the MMLLearner
     * combo box.
     * Method updates selected learner in guimodel.
     */
    private void MMLLearnerComboActionPerformed(java.awt.event.ActionEvent evt) {
    	int idx = MMLLearnerCombo.getSelectedIndex();
    	guimodel.MMLLearner = GUIModel.MMLLearners[idx];
    }
    
    /**User leaves Search Factor text field: Validate input in search factor*/
    private void searchFactorTextfieldFocusLost(java.awt.event.FocusEvent evt) {
    	String s = searchFactorTextfield.getText();
    	if( validateNumericInputStringDouble( s, GUIParameters.minSearchFactor, GUIParameters.maxSearchFactor ) ){
    		searchFactorTextfield.setBackground( Color.WHITE );		//Valid input
    		guimodel.searchFactor = Double.parseDouble(s);
    	} else{
    		searchFactorTextfield.setBackground( Color.RED );		//invalid input
    		guimodel.searchFactor = Double.NaN;
    	}
    }

    /**User leaves maxSECs text field: Validate input*/
    private void maxSECsTextfieldFocusLost(java.awt.event.FocusEvent evt) {
        String s = maxSECsTextfield.getText();
    	if( validateNumericInputStringInteger( s, GUIParameters.minSECs, GUIParameters.maxSECs ) ){
    		maxSECsTextfield.setBackground( Color.WHITE );		//Valid input
    		guimodel.maxSECs = Integer.parseInt(s);
    	} else{
    		maxSECsTextfield.setBackground( Color.RED );		//invalid input
    		guimodel.maxSECs = Integer.MIN_VALUE;
    	}
    }

    /**User leaves Minimum Total Posterior text field: Validate input */
    private void minTotalPosteriorTextFieldFocusLost(java.awt.event.FocusEvent evt) {
    	String s = minTotalPosteriorTextField.getText();
    	if( validateNumericInputStringDouble( s, GUIParameters.min_minTotalPosterior, GUIParameters.max_minTotalPosterior ) ){
    		minTotalPosteriorTextField.setBackground( Color.WHITE );		//Valid input
    		guimodel.minTotalPosterior = Double.parseDouble(s);
    	} else{
    		minTotalPosteriorTextField.setBackground( Color.RED );		//invalid input
    		guimodel.minTotalPosterior = Double.NaN;
    	}
    }
    
    /**User selects a different option in the RNG combo box (in setup tab)*/
    private void setupRNGComboActionPerformed(java.awt.event.ActionEvent evt) {
    	int idx = setupRNGCombo.getSelectedIndex();
        guimodel.r = GUIParameters.RNGs[ idx ];
        guimodel.useSetSeed = GUIParameters.RNGUseSetSeed[ idx ];
        guimodel.useSetSeed2 = GUIParameters.RNGUseSetSeed2[ idx ];
        setupRNGSeed.setEnabled( GUIParameters.RNGUseSetSeed[ idx ] );
        setupRNGSeed2.setEnabled( GUIParameters.RNGUseSetSeed2[ idx ] );
        
        if( !guimodel.useSetSeed ){
        	setupRNGSeed.setBackground( searchParametersPanel.getBackground() );
        } else{
        	setupRNGSeed.setText( Long.toString( guimodel.randomSeed ) );
        }
        
        if( !guimodel.useSetSeed2 ){
        	setupRNGSeed2.setBackground( searchParametersPanel.getBackground() );
        } else{
        	setupRNGSeed2.setText( Long.toString( guimodel.randomSeed2 ) );
        }
    }
    
    /**User exits RNG seed text box: Validate input */
    private void setupRNGSeedFocusLost(java.awt.event.FocusEvent evt) {
    	if( guimodel.useSetSeed == false ){
    		return;
    	}
        String s = setupRNGSeed.getText();
        
        if( validateNumericInputStringInteger( s, Integer.MIN_VALUE, Integer.MAX_VALUE ) ){
    		setupRNGSeed.setBackground( Color.WHITE );		//Valid input
    		guimodel.randomSeed = Integer.parseInt(s);
    	} else{
    		setupRNGSeed.setBackground( Color.RED );		//invalid input
    		guimodel.randomSeed = new Random().nextInt();
    	}
    }

    /**User exits RNG seed2 text box: Validate input*/
    private void setupRNGSeed2FocusLost(java.awt.event.FocusEvent evt) {
    	if( guimodel.useSetSeed2 == false ){
    		return;
    	}
        String s = setupRNGSeed2.getText();
        
        if( validateNumericInputStringLong( s, Integer.MIN_VALUE, Integer.MAX_VALUE ) ){
    		setupRNGSeed2.setBackground( Color.WHITE );		//Valid input
    		guimodel.randomSeed2 = Integer.parseInt(s);
    	} else{
    		setupRNGSeed2.setBackground( Color.RED );		//invalid input
    		guimodel.randomSeed2 = new Random().nextInt();
    	}
    }
    
    /**User checks the "Learn DBN" checkbox (on setup tab) */
    private void setupLearnDBNCheckboxActionPerformed(java.awt.event.ActionEvent evt) {
        //First: Find if checked or unchecked...
    	boolean learnDBNChecked = setupLearnDBNCheckbox.isSelected();
    	if( learnDBNChecked ){
    		//Disable expert priors tab (can't yet use expert priors with DBN code)
    		setExpertPriorsTabEnabled( false );
    		useExpertPriorsCheckbox.setSelected( false );
    		useExpertPriorsCheckbox.setEnabled( false );
    		guimodel.useExpertPriors = false;
    		expertPriorsTextArea.setText("Expert Priors: Not supported when learning DBNs");
    	} else {
    		//Enable expert priors tab...
    		setExpertPriorsTabEnabled( false );
    		useExpertPriorsCheckbox.setSelected( false );
    		useExpertPriorsCheckbox.setEnabled( true );
    		guimodel.useExpertPriors = false;
    		expertPriorsTextArea.setText("");
    	}
    	//Update expert priors status label...
    	updateStatusBarLabelExpertPriors();
    }
    
    
    
    
    
    //***********************************************************
    /*----------------- Expert Priors Tab Actions: --------------*/
    //***********************************************************
    
    /**User checks/unchecks "Use Expert Priors" checkbox in Expert Priors tab*/
    private void useExpertPriorsCheckboxActionPerformed(java.awt.event.ActionEvent evt) {
    	if( useExpertPriorsCheckbox.isSelected() ){
    		setExpertPriorsTabEnabled( true );
    		guimodel.useExpertPriors = true;
    		//Set the text (if expertPriorsTextArea is empty) to default/template text
    		if( expertPriorsTextArea.getText() == null || expertPriorsTextArea.getText().equals("") ){
    			expertPriorsTextArea.setText( GUIModel.defaultNewExpertPriorString );
    		}
    	} else{
    		if( expertPriorsTextArea.getText() != null && expertPriorsTextArea.getText().equals( GUIModel.defaultNewExpertPriorString ) ){
    			expertPriorsTextArea.setText("");
    		}
    		setExpertPriorsTabEnabled( false );
    		guimodel.useExpertPriors = false;
    	}
    	updateStatusBarLabelExpertPriors();
    }
    
    /**User clicks "New" button in Expert Priors Tab: Load the default/template expert priors text
     * to the textbox.*/
    private void newExpertPriorsBtnActionPerformed(java.awt.event.ActionEvent evt) {                                                   
    	expertPriorsTextArea.setText( GUIModel.defaultNewExpertPriorString );
    }                                                  

    /**User clicks the "Load" button in the expert priors tab
     * Allow user to load a set of expert priors from a text file*/
    private void loadExpertPriorsBtnActionPerformed(java.awt.event.ActionEvent evt) {
    	//Set directory:
    	if( guimodel.selectedFile == null ){
    		dataFileChooser.setCurrentDirectory( new File(".") );
    	} else{
    		dataFileChooser.setCurrentDirectory( guimodel.selectedFile );
    	}
    	
    	int result = expertPriorsFileChooser.showOpenDialog(mainTabbedPane);
    	
    	if( result == JFileChooser.APPROVE_OPTION ){	//If user selects a file to open
    		String path = expertPriorsFileChooser.getSelectedFile().getAbsolutePath();
    		
    		BufferedReader inputReader;
    		try{
    			inputReader = new BufferedReader( new FileReader( path ));
    			
    			String fileContents = "";
    			String line;
    			
    			while( (line=inputReader.readLine()) != null ){
    				fileContents += line + "\n";
    			}
    			inputReader.close();
    			
    			expertPriorsTextArea.setText( fileContents );
    			
    		} catch( Exception e ){
    			JOptionPane.showMessageDialog(mainTabbedPane, "Error loading expert prior data - " + e.getMessage() );
    		}
    	}
    }

    /**User clicks the "Save" button in the expert priors tab: Allow user to save
     * the entered priors text to a text file.*/
    private void saveExpertPriorsBtnActionPerformed(java.awt.event.ActionEvent evt) {
    	int result = expertPriorsFileChooser.showSaveDialog(mainTabbedPane);
    	
    	if( result == JFileChooser.APPROVE_OPTION ){	//User enters a location to save file
    		String path = expertPriorsFileChooser.getSelectedFile().getAbsolutePath();
    		
    		BufferedWriter outputWriter;
    		try{
    			outputWriter = new BufferedWriter( new FileWriter( path ));
    			
    			String toWrite = expertPriorsTextArea.getText();
    			
    			outputWriter.write( toWrite );
    			
    			outputWriter.flush();
    			outputWriter.close();
    			
    		} catch( Exception e ){
    			JOptionPane.showMessageDialog(mainTabbedPane, "Error saving expert prior data - " + e.getMessage() );
    		}
    	}
    }

    
    /**User presses "Validate" (i.e. validate expert priors text) in Expert Priors tab.
     * Check if entered text is OK; otherwise, tell the user what's wrong. */
    private void validateExpertPriorsBtnActionPerformed(java.awt.event.ActionEvent evt) {
    	if( !guimodel.dataValid() ){
    		JOptionPane.showMessageDialog(mainTabbedPane, "Data file not selected. Cannot validate expert priors without data file.",
    				"Error",JOptionPane.ERROR_MESSAGE);
    		updateStatusBarLabelExpertPriors();
    		return;
    	}
    	
    	boolean valid = false;
    	try{
    		valid = guimodel.validateExpertPriors( expertPriorsTextArea.getText() );
    	} catch( Exception e ){
    		String exceptionMsg = e.getMessage();
    		
    		//Remove unwanted stack trace text from the exception message... Inelegant, but it works.
    		String removeText = "Exception:";
    		int idx = exceptionMsg.lastIndexOf( removeText );
    		exceptionMsg = exceptionMsg.substring(idx + removeText.length() + 1 );
    		
    		updateStatusBarLabelExpertPriors();
    		JOptionPane.showMessageDialog( mainTabbedPane, "Expert Priors: Invalid\n" + exceptionMsg );	
    	}
    	if( valid ){
    		guimodel.expertPriorsString = expertPriorsTextArea.getText();
    		updateStatusBarLabelExpertPriors();
    		JOptionPane.showMessageDialog( mainTabbedPane, "Expert Priors: OK");
    	}
    }
    
    /**Expert priors text area loses focus: update the status bar label*/
    private void expertPriorsTextAreaFocusLost(java.awt.event.FocusEvent evt) {
        updateStatusBarLabelExpertPriors();
    }
    
    
    
    
    
    //***********************************************************
    /*------------------- Run Tab Actions: --------------------*/
    //***********************************************************

    /**User clicks the "Run" (i.e. run search) button in the Run tab*/
    private void runCammlRunBtnActionPerformed(java.awt.event.ActionEvent evt) {
    	//Check that search is is not currently running:
    	if( searchThread != null && searchThread.isAlive() ){	//Search is currently running!
    		JOptionPane.showMessageDialog( mainTabbedPane, "Cannot start search: Search already underway.");
    		return;
    	}
    	
    	//First: Get/check most recent version of Expert Priors:
    	if( guimodel.useExpertPriors ){
    		try{
    			guimodel.validateExpertPriors( expertPriorsTextArea.getText() );
    			guimodel.expertPriorsString = expertPriorsTextArea.getText();
    		} catch( Exception e ){
    			guimodel.expertPriorsString = null;
    		}
    	}
    	
    	//Set full results table ('results' tab) to null (so results from previous run - if any -
    	// are not displayed while the search is being run.
    	resultsScrollPane.setViewportView( null );
    	
    	//Create and run the search in a separate thread:
		Runnable r = new Runnable(){
			public void run(){
		    	if( checkSearchSettingsAndDisplayErrors() ){
		    		guimodel.runSearch( setupLearnDBNCheckbox.isSelected() );
		    		updateFullResultsDisplay();
		    	}
			}
		};
		
		searchThread = new Thread(r);
		searchThread.start();
    }
    
    /**User clicks "Clear" button in Run tab: Remove all text from console output text area*/
    private void runClearOutputBtnActionPerformed(java.awt.event.ActionEvent evt) {
        runConsoleOutputTextArea.setText( "" );
    }

    
    
    
    
    //***********************************************************
    /*----------------- Results Tab Actions --------------------*/
    //***********************************************************
    
    /**Helper method: Create/update the table with the full results in Results tab*/
    private void updateFullResultsDisplay(){
    	if( guimodel.searchResults != null ){				//Display original CaMML results
    		fullResultsTable = new JTable( new FullResultsTableModel( guimodel.searchResults ) );
    		resultsScrollPane.setViewportView( fullResultsTable );
    	} else if( guimodel.searchResultsDBN != null ){		//Display DBN CaMML results
    		fullResultsTable = new JTable( new FullResultsTableModel( guimodel.searchResultsDBN ) );
    		resultsScrollPane.setViewportView( fullResultsTable );
    	}
    }
    
    /**User clicks the "Export Selected Network" button*/
    private void resultsExportSelectedBNActionPerformed(java.awt.event.ActionEvent evt) {
    	if( guimodel.metropolisSearch == null || !guimodel.metropolisSearch.isFinished() ){
    		JOptionPane.showMessageDialog( mainTabbedPane, "Cannot export network: Search not complete.");
    		return;
    	}
    	if( guimodel.searchResults == null && guimodel.searchResultsDBN == null ){
    		JOptionPane.showMessageDialog( mainTabbedPane, "Cannot export network: Results not generated.");
    		return;
    	}
    	
    	//Find index of selected BN:
    	int rowSelected = fullResultsTable.getSelectedRow();
    	if( rowSelected == -1 ){
    		JOptionPane.showMessageDialog( mainTabbedPane, "Cannot export network: No network Selected.");
    		return;
    	}
        
    	//Get the path to export the BN to (i.e. open the file dialog box) - and export the selected BN:
    	int result = outputBNetChooser.showSaveDialog( this );
    	if( result == JFileChooser.APPROVE_OPTION ){
    		guimodel.lastExportedBNet = outputBNetChooser.getSelectedFile();
    		
    		String outputPath = outputBNetChooser.getSelectedFile().getAbsolutePath();
    		
    		if( !outputPath.endsWith(".dne") ){
    			outputPath += ".dne";
    		}
    		try{
    			guimodel.exportFullResultsBN( outputPath, rowSelected );
    		} catch( Exception e ){
    			JOptionPane.showMessageDialog( mainTabbedPane, "Error during network export.");
    			return;
    		}
    	}
    }
    
    /**User clicks "Export All" button: Export all networks to a specified location*/
    private void resultsExportAllNetworksActionPerformed(java.awt.event.ActionEvent evt) {
    	//Check if full results have actually been generated yet:
    	if( guimodel.metropolisSearch == null || !guimodel.metropolisSearch.isFinished() ){
    		JOptionPane.showMessageDialog( mainTabbedPane, "Cannot Export Networks: Search not complete.");
    		return;
    	}
    	if( (guimodel.searchResults == null && guimodel.searchResultsDBN == null ) || fullResultsTable == null ){
    		JOptionPane.showMessageDialog( mainTabbedPane, "Cannot Export Networks: Full results not generated.");
    		return;
    	}
    	
    	
    	//Set path to current working path, or last directory...
    	if( guimodel.lastExportedBNet == null ){
    		outputBNetChooser.setCurrentDirectory( new File(".") );
    	} else{
    		outputBNetChooser.setCurrentDirectory( guimodel.lastExportedBNet );
    	}
    	
    	//Get output location:
    	int result = outputBNetChooser.showSaveDialog( this );
        
    	if( result == JFileChooser.APPROVE_OPTION ){
    		guimodel.lastExportedBNet = outputBNetChooser.getSelectedFile();
    		
    		String outputPath = outputBNetChooser.getSelectedFile().getAbsolutePath();
    		    		
    		if( !outputPath.endsWith(".dne") ){
    			outputPath += ".dne";
    		}
    		try{
    			guimodel.outputFullResultsAllBNs( outputPath );
    		} catch( Exception e ){
    			JOptionPane.showMessageDialog( mainTabbedPane, "Error during network export.");
    			return;
    		}
    	}
    }
    
    /**User clicks "View Selected Network" button -> Display BN for one of the 'Representative Networks'*/
    private void resultsViewSelectedNetworkBtnActionPerformed(java.awt.event.ActionEvent evt) {
    	//Determine if search run yet
    	if( guimodel.metropolisSearch == null || !guimodel.metropolisSearch.isFinished() ){
    		JOptionPane.showMessageDialog( mainTabbedPane, "Cannot View Network: Search not complete.");
    		return;
    	}
    	if( guimodel.searchResults == null && guimodel.searchResultsDBN == null ){
    		JOptionPane.showMessageDialog( mainTabbedPane, "Cannot View Network: Results not generated.");
    		return;
    	}
    	
    	//Find index of selected row:
    	int rowSelected = fullResultsTable.getSelectedRow();
    	if( rowSelected == -1 ){
    		JOptionPane.showMessageDialog( mainTabbedPane, "Cannot View Network: No network Selected.");
    		return;
    	}
    	
    	//Generate string of selected network:
    	String networkString;
    	try{
    		networkString = guimodel.generateNetworkStringFullResults( rowSelected );
    	} catch( Exception e ){
    		JOptionPane.showMessageDialog( mainTabbedPane, "Error occurred: Cannot view selected network.");
    		return;
    	}
    	
    	//Create viewer:
    	try{
    		String windowTitle = "Representative Network " + (rowSelected+1);
    		new BNetViewer( networkString, false, windowTitle );
        } catch( Exception e ){
        	JOptionPane.showMessageDialog( mainTabbedPane, e );
        }
    }

    
    
    
    //***********************************************************
    /*-------------------- Helper Methods ----------------------*/
    //***********************************************************
    
    /**Helper method: Enable/Disable buttons in the Expert Priors tab
     * (i.e. when user clicks the "Use Expert Priors" checkbox)
     */
    private void setExpertPriorsTabEnabled( boolean enabled ){
    	newExpertPriorsBtn.setEnabled( enabled );
    	loadExpertPriorsBtn.setEnabled( enabled );
    	saveExpertPriorsBtn.setEnabled( enabled );
    	expertPriorsTextArea.setEnabled( enabled );
    	validateExpertPriorsBtn.setEnabled( enabled );
    }
    
    
    
    /** Method to redirect console output to the required place.
     * Currently set up to redirect output to console output to jTextField in the run tab.
     * Other areas of CaMML output important information using System.out.println(...)
     *  hence this is necessary.
     * @param outputTextArea The text area to redirect console output to.
     */
    private void redirectOutputStream( final JTextArea outputTextArea ){
    	//Anonymous inner class for redirecting output stream to the specified text area:
    	OutputStream out = new OutputStream(){
    		@Override
    	    public void write(int b) throws IOException {
    	      outputTextArea.append(String.valueOf((char) b));
    	    }
    	 
    	    @Override
    	    public void write(byte[] b, int off, int len) throws IOException {
    	    	outputTextArea.append(new String(b, off, len));
    	    }
    	 
    	    @Override
    	    public void write(byte[] b) throws IOException {
    	      write(b, 0, b.length);
    	    }
    	};
    	
    	PrintStream PSOut = new PrintStream( out, true );
    	System.setOut( PSOut );
    	System.setErr( PSOut );
    }

    /** Get the options from the GUIParameters and populate the various
     *  combo boxes etc
     */
    private void populateSetupOptions(){
    	if( GUIParameters.MMLLearners.length != GUIParameters.MMLLearnerNames.length ){
    		System.out.println("Error in GUIAvailableParameters code (#Options != #names).");
    		throw new RuntimeException("Error in GUIAvailableParameters code (#Options != #names).");
    	}
    	
    	if( GUIParameters.RNGs.length != GUIParameters.RNGsString.length
    			|| GUIParameters.RNGs.length != GUIParameters.RNGUseSetSeed.length){
    		System.out.println("Error in GUIAvailableParameters code (#Options != #names).");
    		throw new RuntimeException("Error in GUIAvailableParameters code (#Options != #names).");
    	}
    	
    	//Parameterize the available learners in the combo box:
    	MMLLearnerCombo.removeAllItems();
    	
    	for( String s : GUIParameters.MMLLearnerNames ){
    		MMLLearnerCombo.addItem(s);
    	}
    	
    	//Parameterize the available RNGs:
    	setupRNGCombo.removeAllItems();
    	for( String s : GUIParameters.RNGsString ){
    		setupRNGCombo.addItem(s);
    	}
    	//Enable/disable the textfield for the RNG seed if appropriate:
    	setupRNGSeed.setEnabled( GUIParameters.RNGUseSetSeed[0] );
    	setupRNGSeed2.setEnabled( GUIParameters.RNGUseSetSeed2[0] );	
    }
    
    /**Check the selected settings to ensure all values are OK before running.
     * Checks data, MMLLearner and Expert priors, etc.
     * Displays one error (as message dialog) if there is an invalid value for any of these.
     * @return True if search is OK to start. False otherwise.
     */
    private boolean checkSearchSettingsAndDisplayErrors(){
    	//Update status labels:
    	updateStatusBarLabelExpertPriors();
    	updateStatusBarLabelData();
    	
    	//Check data:
    	if( !guimodel.dataValid() ){
    		JOptionPane.showMessageDialog(mainTabbedPane, "Data file not valid/selected.","Error",JOptionPane.ERROR_MESSAGE);
    		return false;
    	}
    	
    	//Check MMLLearner: (note: this should always be OK...)
    	if( !guimodel.MMLLearnerValid() ){
    		JOptionPane.showMessageDialog(mainTabbedPane, "Invalid parameterization type.","Error",JOptionPane.ERROR_MESSAGE);
    		return false;
    	}
    	
    	//Check Search factor value:
    	if( !guimodel.searchFactorValid() ){
    		JOptionPane.showMessageDialog(mainTabbedPane, "Invalid value for Search Factor in Setup.","Error",JOptionPane.ERROR_MESSAGE);
    		return false;
    	}
    	
    	//Check max SECs value:
    	if( !guimodel.maxSECsValid() ){
    		JOptionPane.showMessageDialog(mainTabbedPane, "Invalid value for Maximum Number of SECs in Setup.","Error",JOptionPane.ERROR_MESSAGE);
    		return false;
    	}
    	
    	//Check max SECs value:
    	if( !guimodel.minTotalPosteriorValid() ){
    		JOptionPane.showMessageDialog(mainTabbedPane, "Invalid value for Minimum Total Posterior in Setup.","Error",JOptionPane.ERROR_MESSAGE);
    		return false;
    	}
    	
    	
    	if( !guimodel.expertPriorsValid() ){
    		JOptionPane.showMessageDialog(mainTabbedPane, "Expert priors invalid.","Error",JOptionPane.ERROR_MESSAGE);
    		return false;
    	}
    	
    	return true;
    }
    
    /**Helper method for  the text boxes expecting numeric input (double)
     *Return true if OK number (between minValue and maxValue)*/
    private boolean validateNumericInputStringDouble( String s, double minValue, double maxValue ){
    	if( s == null ) return false;
    	
    	//Try to convert to numeric value:
    	double strVal;
    	try{
    		strVal = Double.parseDouble(s);
    	} catch( NumberFormatException e ){
    		return false;
    	}
    	
    	if( strVal < minValue || strVal > maxValue ) return false;
    	
    	return true;
    }
    
    /**Helper method for  the text boxes expecting numeric input (integer)
     *Return true if OK number (between minValue and maxValue)*/
    private boolean validateNumericInputStringInteger( String s, int minValue, int maxValue ){
    	if( s == null ) return false;
    	
    	//Try to convert to numeric value:
    	int strVal;
    	try{
    		strVal = Integer.parseInt(s);
    	} catch( NumberFormatException e ){
    		return false;
    	}
    	
    	if( strVal < minValue || strVal > maxValue ) return false;
    	
    	return true;
    }
    
    /**Helper method for  the text boxes expecting numeric input (long)
     *Return true if OK number (between minValue and maxValue)*/
    private boolean validateNumericInputStringLong( String s, long minValue, long maxValue ){
    	if( s == null ) return false;
    	
    	//Try to convert to numeric value:
    	long strVal;
    	try{
    		strVal = Long.parseLong(s);
    	} catch( NumberFormatException e ){
    		return false;
    	}
    	
    	if( strVal < minValue || strVal > maxValue ) return false;
    	
    	return true;
    }
    
    /** Generates a JFrame for viewing the data in the selected file.
     */
    private void createDataViewer(){
    	if( guimodel.dataValid() ){
			JFrame frame = new DataViewer(guimodel.data);
			frame.setSize(800,600);
			frame.setTitle( "Data Viewer - " + guimodel.selectedFile.getName() );
			frame.setDefaultCloseOperation( JFrame.DISPOSE_ON_CLOSE );
			frame.setVisible(true);
    	}
    }
    
    /**Update the status bar label for Data*/
    private void updateStatusBarLabelData(){
    	if( guimodel.data == null ){
    		statusBarLabelData.setText("Data: Not Loaded");
    		return;
    	}
    	if( !guimodel.dataValid() ){
    		statusBarLabelData.setText("Data: Not Valid");
    		return;
    	}
    	//Data must be valid...
    	statusBarLabelData.setText("Data: Loaded, OK");
    }
    
    /**Update the status bar label for Expert Priors*/
    private void updateStatusBarLabelExpertPriors(){
    	if( setupLearnDBNCheckbox.isSelected() ){
    		statusBarLabelExpertPriors.setText("Expert Priors: Not Supported for DBNs");
    		return;
    	}
    	if( guimodel.useExpertPriors == false ){
    		statusBarLabelExpertPriors.setText( "Expert Priors: Not Set");
    	} else{
    		if( expertPriorsTextArea.getText().equals( guimodel.expertPriorsString ) ){
    			statusBarLabelExpertPriors.setText("Expert Priors: Set, Validated");
    		} else{
    			statusBarLabelExpertPriors.setText("Expert Priors: Set, Not Validated");
    		}
    	}
    }
    

    
    //=========================================================
    // ----- Variable Declaration: Generated by NetBeans -----
    private javax.swing.JComboBox MMLLearnerCombo;
    private javax.swing.JMenuItem aboutMenuItem;
    private javax.swing.JMenuItem exitMenuItem;
    private javax.swing.JTextArea expertPriorsTextArea;
    private javax.swing.JMenu fileMenu;
    private javax.swing.JTextField filePathTextField;
    private javax.swing.JMenu helpMenu;
    private javax.swing.JMenuItem helpMenuItem;
    private javax.swing.JLabel jLabelDataFileLabel;
    private javax.swing.JMenuBar jMenuBar1;
    private javax.swing.JScrollPane jScrollPane1;
    private javax.swing.JScrollPane jScrollPane2;
    private javax.swing.JPopupMenu.Separator jSeparator2;
    private javax.swing.JSeparator jSeparator3;
    private javax.swing.JSeparator jSeparator4;
    private javax.swing.JButton loadDataFileBtn;
    private javax.swing.JButton loadExpertPriorsBtn;
    private javax.swing.JTabbedPane mainTabbedPane;
    private javax.swing.JTextField maxSECsTextfield;
    private javax.swing.JTextField minTotalPosteriorTextField;
    private javax.swing.JButton newExpertPriorsBtn;
    private javax.swing.JPanel priorsPanel;
    private javax.swing.JButton resultsExportAllNetworks;
    private javax.swing.JButton resultsExportSelectedBN;
    private javax.swing.JLabel resultsLabelFullResults;
    private javax.swing.JPanel resultsPanel;
    private javax.swing.JScrollPane resultsScrollPane;
    private javax.swing.JButton resultsViewSelectedNetworkBtn;
    private javax.swing.JButton runCammlRunBtn;
    private javax.swing.JButton runClearOutputBtn;
    private javax.swing.JTextArea runConsoleOutputTextArea;
    private javax.swing.JLabel runLabelConsoleOutput;
    private javax.swing.JPanel runPanel;
    private javax.swing.JButton saveExpertPriorsBtn;
    private javax.swing.JTextField searchFactorTextfield;
    private javax.swing.JPanel searchParametersPanel;
    private javax.swing.JScrollPane searchParametersScrollPane;
    private javax.swing.JLabel setupLabelAdvancedSettings;
    private javax.swing.JLabel setupLabelLearnDBN;
    private javax.swing.JLabel setupLabelMMLLearner;
    private javax.swing.JLabel setupLabelMaxSECs;
    private javax.swing.JLabel setupLabelMinTotalPosterior;
    private javax.swing.JLabel setupLabelParameterizationSettings;
    private javax.swing.JLabel setupLabelRNG;
    private javax.swing.JLabel setupLabelRNGSeed;
    private javax.swing.JLabel setupLabelSearchFactor;
    private javax.swing.JLabel setupLabelSearchSettings;
    private javax.swing.JCheckBox setupLearnDBNCheckbox;
    private javax.swing.JPanel setupPanel;
    private javax.swing.JComboBox setupRNGCombo;
    private javax.swing.JTextField setupRNGSeed;
    private javax.swing.JTextField setupRNGSeed2;
    private javax.swing.JLabel statusBarLabelData;
    private javax.swing.JLabel statusBarLabelExpertPriors;
    private javax.swing.JCheckBox useExpertPriorsCheckbox;
    private javax.swing.JButton validateExpertPriorsBtn;
    private javax.swing.JLabel versionLabel;
    private javax.swing.JButton viewDataBtn;
    //End of variable declaration - NetBeans autogenerated...
    
    //For loading data files, saving networks, loading expert priors:
    private JFileChooser dataFileChooser;
    private JFileChooser expertPriorsFileChooser;
    private JFileChooser outputBNetChooser;
    
    //Full results table:
    private JTable fullResultsTable;
    
    //Separate Thread for running search in:
    Thread searchThread = null;			//Thread for running the search
}



